#define _GNU_SOURCE
#include "activity.h"
#include "activity/context.h"

#include "ffdb/fs_list.h"
#include "collections/array.h"
#include "format.h"

#include "model/server.h"
#include "model/account.h"
#include "model/status.h"
#include "model/ap/outbox_envelope.h"
#include "model/ap/activity/rsa_signature_2017.h"

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

struct ap_activity* ap_activity_new()
{
	struct ap_activity* act = malloc(sizeof(struct ap_activity));
	memset(act,0,sizeof(struct ap_activity));

	return act;
}
struct ap_activity* ap_activity_dup( struct ap_activity* act )
{
	struct ap_activity* new_act = ap_activity_new();
	memset(new_act,0,sizeof(*new_act));

	new_act->id = strdup(act->id);
	new_act->local_id = act->local_id;
	new_act->type = act->type;
	new_act->actor = strdup(act->actor);
	new_act->attributed_to = safe_strdup(act->attributed_to);
	new_act->published = act->published;

	new_act->in_reply_to = safe_strdup( act->in_reply_to );

	new_act->source.content = safe_strdup(act->source.content);
	new_act->source.mime_type = safe_strdup(act->source.mime_type);

	new_act->content.content = safe_strdup(act->content.content);
	new_act->content.mime_type = safe_strdup(act->content.mime_type);

	array_dup( &new_act->to, sizeof(char*), &act->to );
	for( int i = 0; i < new_act->to.count; ++i ) {
		new_act->to.items[i] = strdup(new_act->to.items[i]);
	}

	array_dup( &new_act->cc, sizeof(char*), &act->cc );
	for( int i = 0; i < new_act->cc.count; ++i ) {
		new_act->cc.items[i] = strdup(new_act->cc.items[i]);
	}

	array_dup( &new_act->bcc, sizeof(char*), &act->bcc );
	for( int i = 0; i < new_act->bcc.count; ++i ) {
		new_act->bcc.items[i] = strdup(new_act->bcc.items[i]);
	}

	switch( act->object.tag ) {
		case apaot_ref:
			new_act->object.ref = strdup(act->object.ref);
			break;
		case apaot_activity:
			new_act->object.ptr = ap_activity_dup(act->object.ptr);
			break;
	}
	new_act->object.tag = act->object.tag;

	new_act->state = safe_strdup( act->state );

	if( act->has_signature ) {
		new_act->has_signature = true;
		new_act->signature.type = act->signature.type;
		new_act->signature.creator = strdup(act->signature.creator);
		new_act->signature.created = act->signature.created;
		new_act->signature.value = strdup(act->signature.value);
	}

	return new_act;
}
void ap_activity_free( struct ap_activity* act )
{
	if( !act ) { return; }
	ap_activity_free_composite(act);
	free(act);
}
void ap_activity_free_composite( struct ap_activity* act )
{
	free(act->id);
	free(act->actor);
	free(act->context);
	free(act->attributed_to);
	free(act->target);
	free(act->content.content);
	free(act->source.content);
	free(act->conversation);
	free(act->summary);
	free(act->in_reply_to);

	for( int i = 0; i < act->tags.count; ++i ) {
		ap_activity_tag_free(act->tags.items[i]);
	}
	free(act->tags.items);

	for( int i = 0; i < act->also_known_as.count; ++i ) {
		free( act->also_known_as.items[i] );
	}
	free(act->also_known_as.items);

	for( int i = 0; i < act->attachments.count; ++i ) {
		ap_attachement_free( act->attachments.items[i] );
	}
	free( act->attachments.items );

	for( int i = 0; i < act->to.count; ++i ) {
		free(act->to.items[i]);
	}
	free(act->to.items);

	for( int i = 0; i < act->cc.count; ++i ) {
		free(act->cc.items[i]);
	}
	free(act->cc.items);

	for( int i = 0; i < act->bcc.count; ++i ) {
		free(act->bcc.items[i]);
	}
	free(act->bcc.items);

	switch( act->object.tag ) {
		case apaot_ref:
			free( act->object.ref );
			break;
		case apaot_activity:
			ap_activity_free( act->object.ptr );
			act->object.ptr = 0;
			break;
	};
	free( act->state );

	free( act->signature.creator );
	free( act->signature.value );
}

struct ap_activity* ap_activity_create_accept( struct ap_activity* act )
{
	int id = fs_list_get("data/activities/HEAD") + 1;
	fs_list_set( "data/activities/HEAD", id );

	struct ap_activity* accept = malloc(sizeof(struct ap_activity));
	memset(accept,0,sizeof(*accept));

	asprintf( &accept->id,"https://%s/activity/%d", g_server_name, id );
	accept->local_id = id;
	accept->type = apat_accept;
	asprintf( &accept->actor, "https://%s/owner/actor", g_server_name );
	char* new_act_actor = strdup(act->actor);
	array_append( &accept->to, sizeof(char*), &new_act_actor );
	accept->object.tag = apaot_activity;
	accept->object.ptr = ap_activity_dup(act);

	return accept;
}
void ap_activity_accept( struct ap_activity* act, int deliver_to_account_id )
{
	struct ap_activity* accept = ap_activity_create_accept(act);
	ap_activity_save(accept);

	struct outbox_envelope* env = outbox_envelope_new();
		env->activity_id = accept->local_id;
		env->account_id = deliver_to_account_id;
	outbox_envelope_save( env );
	outbox_envelope_free( env );

	ap_activity_free(accept);
}

struct ap_activity* ap_activity_create_emoji_react( struct status* s, const char* react )
{
	struct account* a = account_from_id( s->account_id );

	int id = fs_list_get("data/activities/HEAD") + 1;
	fs_list_set( "data/activities/HEAD", id );

	struct ap_activity* act = ap_activity_new();
	act->local_id = id;
	act->id = aformat( "https://%s/activity/%d", g_server_name, id );
	act->actor = aformat( "https://%s/owner/actor", g_server_name );
	act->type = apat_emoji_react;
	act->content.content = safe_strdup(react);
	act->published = time(NULL);
	act->object.tag = apaot_ref;
	act->object.ref = strdup( s->url );

	char* to = aformat( "%s/followers", a->account_url );
	array_append( &act->to, sizeof(to), &to );

	char* cc = strdup( "https://www.w3.org/ns/activitystreams#Public" );
	array_append( &act->cc, sizeof(cc), &cc );
	cc = strdup( a->account_url );
	array_append( &act->cc, sizeof(cc), &cc );

	account_free(a);
	return act;
}

void ap_activity_react( struct status* s, const char* react )
{
	struct ap_activity* act = ap_activity_create_emoji_react(s,react);
	ap_activity_save(act);

	struct outbox_envelope* env = outbox_envelope_new();
		env->activity_id = act->local_id;
		env->account_id = s->account_id;
	outbox_envelope_save( env );
	outbox_envelope_free( env );

	ap_activity_free(act);
}

struct ap_activity* ap_activity_create_follow( struct account* follower, struct account* following )
{
	int id = fs_list_get("data/activities/HEAD") + 1;
	fs_list_set( "data/activities/HEAD", id );

	struct ap_activity* act = ap_activity_new();
	act->local_id = id;
	act->id = aformat( "https://%s/activity/%d", g_server_name, id );
	act->actor = strdup( follower->account_url );
	act->type = apat_follow;
	act->published = time(NULL);
	act->object.tag = apaot_ref;
	act->state = strdup("pending");
	act->object.ref = strdup( following->account_url );

	char* to = strdup(following->account_url);
	array_append( &act->to, sizeof(to), &to );

	return act;
}
int ap_activity_follow( struct account* follower, struct account* following )
{
	int res = -1;

	struct ap_activity* act = ap_activity_create_follow( follower, following );
	if( !act ) { goto failed; }

	ap_activity_save(act);
	res = act->local_id;

	struct outbox_envelope* env = outbox_envelope_new();
	if( !env ) { goto failed; }
		env->activity_id = act->local_id;
		env->account_id = following->id;
	outbox_envelope_save( env );
	outbox_envelope_free( env );
cleanup:
	ap_activity_free(act);
	return res;
failed:
	res = -1;
	goto cleanup;
}

struct ap_activity* ap_activity_create_undo( struct ap_activity* act_to_undo )
{
	int id = fs_list_get("data/activities/HEAD") + 1;
	fs_list_set( "data/activities/HEAD", id );

	struct ap_activity* act = ap_activity_new();
	act->local_id = id;
	act->id = aformat( "https://%s/activity/%d", g_server_name, id );
	act->actor = strdup( act_to_undo->actor );
	act->type = apat_undo;
	act->published = time(NULL);
	act->object.tag = apaot_activity;
	act->object.ptr = ap_activity_dup(act_to_undo);

	char* to = strdup(act_to_undo->object.ref);
	array_append( &act->to, sizeof(to), &to );

	return act;
}

void ap_activity_undo( struct ap_activity* act, int deliver_to_account_id )
{
	struct ap_activity* undo_act = ap_activity_create_undo( act );
	if( !undo_act ) { goto failed; }

	ap_activity_save(undo_act);
	/*
	printf( "act=" );
	ap_activity_write_to_FILE( act, stdout );
	printf( "\nundo_act=");
	ap_activity_write_to_FILE( undo_act, stdout );
	printf( "\n" );
	*/

	struct outbox_envelope* env = outbox_envelope_new();
	if( !env ) { goto failed; }
		env->activity_id = undo_act->local_id;
		env->account_id = deliver_to_account_id;
	outbox_envelope_save( env );
	outbox_envelope_free( env );
cleanup:
	ap_activity_free(undo_act);
	return;
failed:
	goto cleanup;
}

struct ap_activity* ap_activity_create_like( struct status* s )
{
}

int ap_activity_like( struct status* s )
{
	printf( "TODO: like\n" );

	return -1;
}

struct ap_activity* ap_activity_create_note( struct status* s )
{
	struct ap_activity* act = ap_activity_new();
	struct account* a = account_from_id( s->account_id );

	act->id = aformat( "https://%s/note/%d", g_server_name, s->id );
	act->type = apat_note;
	act->actor = strdup( a->account_url );
	act->attributed_to = strdup( a->account_url );
	act->published = s->published;
	act->source.content = strdup(s->source);
	act->content.content = strdup(status_render_source(s));
	if( s->in_reply_to ) {
		struct status* s_in_reply_to = status_from_id( s->in_reply_to );
		act->in_reply_to = strdup( s_in_reply_to->url );
		status_free(s_in_reply_to);
	}
	char* str = strdup("https://www.w3.org/ns/activitystreams#Public");
	array_append( &act->to, sizeof(str), &str );

	str = aformat( "https://%s/owner/followers", g_server_name );
	array_append( &act->cc, sizeof(str), &str );

	for( int i = 0; i < s->mentions.count; ++i ) {
		struct account* mentioned = account_from_id( s->mentions.items[i] );

		str = strdup(mentioned->account_url);
		array_append( &act->to, sizeof(str), &str );

		struct ap_activity_tag* tag;
		tag = malloc(sizeof(*tag));
		tag->type = aptag_mention;
		tag->href = strdup(mentioned->account_url);
		tag->name = aformat( "%s@%s", mentioned->handle, mentioned->server );
		array_append( &act->tags, sizeof(tag), &tag );
	}

	account_free(a);
	return act;
}

struct ap_activity* ap_activity_create_Create( struct ap_activity* object )
{
	int id = fs_list_get("data/activities/HEAD") + 1;
	fs_list_set( "data/activities/HEAD", id );

	struct ap_activity* act = ap_activity_new();
	act->local_id = id;
	act->type = apat_create;
	act->id = aformat( "https://%s/activity/%d", g_server_name, id );
	act->object.tag = apaot_activity;
	act->object.ptr = ap_activity_dup(object);
	act->actor = strdup( object->actor );
	act->published = object->published;

	char* str;
	for( int i = 0; i < object->to.count; ++i ) {
		str = strdup( object->to.items[i] );
		array_append( &act->to, sizeof(str), &str );
	}
	for( int i = 0; i < object->cc.count; ++i ) {
		str = strdup( object->cc.items[i] );
		array_append( &act->cc, sizeof(str), &str );
	}

	return act;
}


